# noinspection PyUnresolvedReferences
# - Convenience pass-through from "private" sub-module.
from ._defaults import DFLT_CLASSIFIER_FACTORY, DFLT_DETECTION_FACTORY
from ._interface import ObjectDetector, ImageMatrixObjectDetector
